/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 
**      10        20        30        40        50        60        70        80
**
** program:
**    cairo-dock
**
** author(s):
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**    Behdad Esfahbod <behdad@behdad.org>
**    David Reveman <davidr@novell.com>
**    Karl Lattimer <karl@qdh.org.uk> 
**
** Copyright (C) Mirco Mueller, June 2006.
**
** notes:
**    A stress-test for cairo, librsvg, and glitz.  This program is released
**    under the terms of the GNU General Public License. If you don't know what
**    that means take a look at:
**
**        http://www.gnu.org/licenses/licenses.html#GPL
**
** notes from original author:
**
**    I just know that some folks will bug me regarding this, so... yes there's
**    nearly everything hard-coded, it is not nice, it is not very usable for
**    easily (without any hard work) making a full dock-like application out of
**    this, please don't email me asking to me to do so... for everybody else
**    feel free to make use of it beyond this being a small but heavy stress
**    test. I've written this on an Ubuntu-6.06 box running Xgl/compiz. The
**    icons used are from the tango-project...
**
**        http://tango-project.org/
**
**    Over the last couple of days Behdad and David helped me (MacSlow) out a
**    great deal by sending me additional tweaked and optimized versions. I've
**    now merged all that with my recent additions.
**
*******************************************************************************/

#include <math.h>
#include <string.h>
#include <cairo.h>
#include <pango/pango.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <librsvg/rsvg.h>
#include <librsvg/rsvg-cairo.h>
#ifdef HAVE_GLITZ
#include <gdk/gdkx.h>
#include <glitz-glx.h>
#include <cairo-glitz.h>
#endif

#include "tango-colors.h"

#define WIN_HEIGHT 150
#define INFLUENCE 150.0
#define GAP 1.0

typedef struct _Icon {
	gchar acFileName[64];
	gchar acName[32];
	gchar acCommand[90];
	gdouble fX;
	gdouble fY;
	gdouble fWidth;
	gdouble fHeight;
	gdouble fScale;
	gdouble fImageScale;
	cairo_surface_t* pIconBuffer;
	cairo_surface_t* pTextBuffer;
	gdouble fTextXOffset;
	gdouble fTextYOffset;
	RsvgHandle* pHandle;
	gboolean bClicked;
	gdouble fPeriod;
} Icon;

static Icon g_aIcons[] =
{
	{"user-home.svg", "Home", "nautilus /home/guillaume"},
	{"gnome-terminal.svg", "Terminal", "gnome-terminal --working-directory=/home/guillaume"},
	{"mozilla-firefox.svg", "Firefox", "firefox"},
	{"amarok.svg", "AmaroK", "amarok"},
	{"gaim.svg", "Gaim", "gaim"},
	{"gxine.svg", "Totem", "totem"},
	{"ooo-writer.svg", "Writer", "ooffice -writer"},
	{"ooo-calc.svg", "Calc", "ooffice -calc"},
	{"ooo-impress.svg", "Impress", "ooffice -impress"},
	{"xmms.svg", "Sound Juicer", "sound-juicer"},
	{"gnome-gimp.svg", "GIMP", "gimp"},
	{"azureus.svg", "Azureus", "/opt/azureus/azureus"},
	{"dc++.svg", "DC++", "/opt/linuxdcpp/ldcpp"},
	{"gnome-calculator.svg", "Calculator", "gcalctool"},
	{"beagled.svg", "Beagle", "beagle-search"},
	{"gconf-editor.svg", "Configuration", "gconf-editor"},

};

#define SVGS ((int)(sizeof (g_aIcons) / sizeof (g_aIcons[0])))

static RsvgDimensionData g_DimensionData;
static gint g_iScreenHeight = 0;
static gint g_iCurrentWidth = 0;
static gint g_iCurrentHeight = WIN_HEIGHT;
static gboolean g_bClicked = FALSE;
static gint g_iClickedIcon = -1;
static float g_fClickedIconX = 0.0;
static float g_fMouseX = 0.0;
static float g_fMouseY = 0.0;
static guint g_uiRedraws = 0;
static guint g_uiEvents = 0;
static guint g_uiTimeoutHandlerId = 0;
static guint g_uiHideHandlerId = 0;
static float g_fInfluence = 0.0;
static float g_fMagnitude = 0.0;

#define UP 1
#define DOWN 2
#define STATIC 0

static gint g_iWindowMove = STATIC;
static gint g_iAtBottom = FALSE;

static gint g_iWindowPositionX = 0; /* dock-windows current y-position */
static gint g_iWindowPositionY = 0; /* dock-windows current y-position */
static gint g_iWindowWidth = 0; /* full screen-width */
static gint g_iWindowHeight = 0; /* 3 * icon-height */
static gint g_iDockLineWidth = 2; /* thickness of dock-bg outline */
static gint g_iDockRadius = 15; /* radius of dock-bg corners */
static gint g_iDockWidth = 0; /* current dock-width */
static gint g_iDockHeight = 0; /* current dock-height */
static gint g_iDockOffsetX = 0; /* x-position for unscaled dock */
static gint g_iDockOffsetY = 0; /* y-position for unscaled dock */
static gdouble g_fGradientOffsetX = 4.0f; /* controls gradient-animation */
static gint iInfluenceGlobal;

static gboolean g_bDrawBackground = TRUE;
static gboolean g_bUseIconBuffering = TRUE;
static gboolean g_bUseText = TRUE;
static gboolean g_bKeepAbove = TRUE;
static gboolean g_bSkipPager = TRUE;
static gboolean g_bSkipTaskbar = TRUE;
static gboolean g_bSticky = TRUE;

#ifdef HAVE_GLITZ
static gboolean g_bUseGlitz = TRUE;
static glitz_drawable_format_t *gDrawFormat = NULL;
static glitz_drawable_t* g_pGlitzDrawable = NULL;
static glitz_format_t* g_pGlitzFormat = NULL;
#endif

static gboolean g_bUseBenchmark = FALSE;
static int g_iAutoMoveDir = 1;

static void recalc_positions (Icon* pIcon, int iIcons, gdouble fX, gdouble fGap, gdouble fInfluence, gdouble fMagnitude);
static gboolean shrink_down (gpointer data);
static gboolean grow_up (gpointer data);
static gboolean clicked (gpointer data);
static gboolean none_clicked ();

static cairo_t *
dock_cairo_create (GdkWindow* pWindow)
{
#ifdef HAVE_GLITZ
	if (g_pGlitzDrawable)
	{
		glitz_surface_t* pGlitzSurface;
		cairo_surface_t* pCairoSurface;
		cairo_t* pCairoContext;

		pGlitzSurface = glitz_surface_create (g_pGlitzDrawable,
						      g_pGlitzFormat,
						      g_iCurrentWidth,
						      g_iCurrentHeight,
						      0,
						      NULL);

		if (gDrawFormat->doublebuffer)
			glitz_surface_attach (pGlitzSurface,
					      g_pGlitzDrawable,
					      GLITZ_DRAWABLE_BUFFER_BACK_COLOR);
		else
			glitz_surface_attach (pGlitzSurface,
					      g_pGlitzDrawable,
					      GLITZ_DRAWABLE_BUFFER_FRONT_COLOR);

		pCairoSurface = cairo_glitz_surface_create (pGlitzSurface);
		pCairoContext = cairo_create (pCairoSurface);

		cairo_surface_destroy (pCairoSurface);
		glitz_surface_destroy (pGlitzSurface);

		return pCairoContext;
	}
#endif

	return gdk_cairo_create (pWindow);
}

static void
on_alpha_screen_changed (GtkWidget* pWidget,
			 GdkScreen* pOldScreen,
			 GtkWidget* pLabel)
{                       
	GdkScreen* pScreen = gtk_widget_get_screen (pWidget);
	GdkColormap* pColormap = gdk_screen_get_rgba_colormap (pScreen);
      
	if (!pColormap)
		pColormap = gdk_screen_get_rgb_colormap (pScreen);

	gtk_widget_set_colormap (pWidget, pColormap);
}

static gint
get_current_dock_width ()
{
	gint iWidth = 0;

	iWidth = g_aIcons[SVGS-1].fX + g_aIcons[SVGS-1].fWidth * g_aIcons[SVGS-1].fScale - g_aIcons[0].fX + 2 * g_iDockRadius + g_iDockLineWidth;

	return iWidth;
}

static gint
get_dock_offset_y ()
{
	return g_iWindowHeight - g_aIcons[0].fHeight - g_iDockLineWidth;
}

static gint
get_current_dock_offset_x ()
{
	return g_aIcons[0].fX;
}

static gint
get_current_dock_offset_y ()
{
	return g_iWindowHeight - g_aIcons[0].fHeight - g_iDockLineWidth;
}

static void
render (cairo_t* pCairoContext,
	gint iWidth,
	gint iHeight)
{
	int icon;
	cairo_pattern_t* pPattern = NULL;

	/* set rendering-"fidelity" and clear canvas */
	cairo_set_tolerance (pCairoContext, 0.1);
	cairo_set_source_rgba (pCairoContext, 0.0, 0.0, 0.0, 0.0);
	cairo_set_operator (pCairoContext, CAIRO_OPERATOR_SOURCE);
	cairo_paint (pCairoContext);

	/* draw background */
	if (g_bDrawBackground)
	{
		cairo_save (pCairoContext);
		pPattern = cairo_pattern_create_linear (0.0f,
												0.0f,
												100.0f,
												(gdouble) iHeight);
		if (cairo_pattern_status (pPattern) == CAIRO_STATUS_SUCCESS)
		{
			gdouble fGradientAlpha = 0.4f;
			cairo_matrix_t matrix;
			gint iRadius = g_iDockRadius;
			gint iLineWidth = g_iDockLineWidth;
			gint iDockWidth = get_current_dock_width ();
			gint iDockOffsetX = get_current_dock_offset_x ();
			gint iDockOffsetY = get_dock_offset_y ();
			gdouble fStep;

			cairo_pattern_set_extend (pPattern, CAIRO_EXTEND_REPEAT);
			cairo_matrix_init_translate (&matrix, g_fGradientOffsetX, 10.0f);
			cairo_pattern_set_matrix (pPattern, &matrix);
			for (fStep = 0.0f; fStep < 1.0f; fStep += 0.1f)
			{
				cairo_pattern_add_color_stop_rgba (pPattern,
												   fStep,
												   TANGO_COLOR_ALUMINIUM1_LIGHT,
												   fGradientAlpha);
				cairo_pattern_add_color_stop_rgba (pPattern,
												   fStep + 0.025f,
												   TANGO_COLOR_ALUMINIUM1_MID,
												   fGradientAlpha);
				cairo_pattern_add_color_stop_rgba (pPattern,
												   fStep + 0.05f,
												   TANGO_COLOR_ALUMINIUM1_LIGHT,
												   fGradientAlpha);
				cairo_pattern_add_color_stop_rgba (pPattern,
												   fStep + 0.075f,
												   TANGO_COLOR_ALUMINIUM1_MID,
												   fGradientAlpha);
			}
			cairo_pattern_add_color_stop_rgba (pPattern,
											   1.0f,
											   TANGO_COLOR_ALUMINIUM1_LIGHT,
											   fGradientAlpha);

			cairo_move_to (pCairoContext, iDockOffsetX, iDockOffsetY - iLineWidth/2);
			
			cairo_rel_line_to (pCairoContext, iDockWidth -2 * iRadius - iLineWidth, 0);
			/* Top Right */
			cairo_rel_curve_to (pCairoContext,
								0, 0,
								iRadius, 0,
								iRadius, iRadius);
			cairo_rel_line_to (pCairoContext,
							   0, iHeight - iDockOffsetY - iRadius);/// - 2 * iRadius);
			/* Bottom Right */
			/*cairo_rel_curve_to (pCairoContext,
								0, 0,
								0, iRadius,
								-iRadius, iRadius);*/

			cairo_rel_line_to (pCairoContext,
							   -iDockWidth + iLineWidth, 0);
			/* Bottom Left */
			/*cairo_rel_curve_to (pCairoContext,
								0, 0,
								-iRadius, 0,
								-iRadius, -iRadius);
			cairo_rel_line_to (pCairoContext, -iRadius, -iRadius);*/


			cairo_rel_line_to (pCairoContext,
							   0, -iHeight + iDockOffsetY + iRadius);
			/* Top Left */
			cairo_rel_curve_to (pCairoContext,
								0, 0,
								0, -iRadius,
								iRadius, -iRadius);

			cairo_set_source (pCairoContext, pPattern);
			cairo_fill_preserve (pCairoContext);
			cairo_set_line_width (pCairoContext, iLineWidth);
			cairo_set_source_rgba (pCairoContext,
								   TANGO_BLACK,
								   0.3f);
			cairo_stroke (pCairoContext);
			cairo_pattern_destroy (pPattern);
		}
		cairo_restore (pCairoContext);
	}

	cairo_set_operator (pCairoContext, CAIRO_OPERATOR_OVER);

	for (icon = 0; icon < SVGS; icon++)
	{
		cairo_save (pCairoContext);
		cairo_translate (pCairoContext, g_aIcons[icon].fX, g_aIcons[icon].fY);


		/* draw the icon */
		cairo_save (pCairoContext);
		if (g_aIcons[icon].pIconBuffer)
		{
			cairo_scale (pCairoContext, g_aIcons[icon].fScale / 2.0, g_aIcons[icon].fScale / 2.0);
			cairo_set_source_surface (pCairoContext, g_aIcons[icon].pIconBuffer, 0.0, 0.0);
			cairo_paint (pCairoContext);
		}
		else
		{
			cairo_scale (pCairoContext, g_aIcons[icon].fScale, g_aIcons[icon].fScale);
			rsvg_handle_render_cairo (g_aIcons[icon].pHandle, pCairoContext);
		}
		cairo_restore (pCairoContext);

		/* draw the label-text */
		if (g_aIcons[icon].pTextBuffer && g_aIcons[icon].fScale > 1.0)
		{
			cairo_set_source_surface (pCairoContext, g_aIcons[icon].pTextBuffer,
						  -g_aIcons[icon].fTextXOffset + g_aIcons[icon].fWidth * g_aIcons[icon].fScale * 0.5,
						  -g_aIcons[icon].fTextYOffset + 0.0);
			cairo_paint_with_alpha (pCairoContext, (g_aIcons[icon].fScale *3) - 5.0 );
		}
		// Made this tighter [KL]
		cairo_restore (pCairoContext);
	}

#ifdef HAVE_GLITZ
	if (gDrawFormat && gDrawFormat->doublebuffer)
		glitz_drawable_swap_buffers (g_pGlitzDrawable);
#endif

	g_uiRedraws++;
}

static gboolean
move_up (gpointer pWidget)
{
	guint distance;
	guint max_dist;
	guint start;
	gint move;
	
	recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);
	gtk_widget_queue_draw ((GtkWidget*) pWidget);

	if (g_uiHideHandlerId == 0) {
		g_iWindowMove = STATIC;
		return FALSE;
	}
	if (g_iWindowPositionY > g_iScreenHeight - WIN_HEIGHT)
	{
		g_iAtBottom = FALSE;

		max_dist = g_iScreenHeight - WIN_HEIGHT;
		start = g_aIcons[0].fHeight + 14;
		distance = g_iWindowPositionY - max_dist;
		move = (distance/6)+1;

		//g_print ("dist: %d, max: %d\n", distance, g_iWindowPositionY);

		
		g_iWindowPositionY -= move;
		gtk_window_move (GTK_WINDOW (pWidget),
						 g_iWindowPositionX,
						 g_iWindowPositionY);
		return TRUE;
	} else {
		if (g_uiHideHandlerId != 0)
		{
			g_source_remove (g_uiHideHandlerId);
			g_uiHideHandlerId = 0;
		}
		g_iWindowMove = STATIC;
		return FALSE;
	}
}

static gboolean
move_down (gpointer pWidget)
{
	guint distance;
	guint max_dist;
	guint start;
	gint move;
	
	recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);
	gtk_widget_queue_draw ((GtkWidget*) pWidget);

	if (g_uiHideHandlerId == 0) {
		g_iWindowMove = STATIC;
		return FALSE;
	}

	if (g_iWindowPositionY < g_iScreenHeight - (WIN_HEIGHT - g_aIcons[0].fHeight - 14))
	{
		max_dist = g_iScreenHeight - WIN_HEIGHT;
		start = g_aIcons[0].fHeight + 14;
		distance = g_iWindowPositionY - max_dist;
		move = (distance/7)+1;

		//g_print ("dist: %d, max: %d\n", move, start);

		g_iWindowPositionY += move;

		if (g_iWindowPositionY > g_iScreenHeight - (WIN_HEIGHT - g_aIcons[0].fHeight - 14))
			g_iWindowPositionY = g_iScreenHeight - (WIN_HEIGHT - g_aIcons[0].fHeight - 14);

		gtk_window_move (GTK_WINDOW (pWidget),
						 g_iWindowPositionX,
						 g_iWindowPositionY);
		return TRUE;
	} else {
		// Should jump off screen -3px when we hit the end
		g_iAtBottom = TRUE;

		gtk_window_move (GTK_WINDOW (pWidget),
						 g_iWindowPositionX,
						 g_iScreenHeight - 3);

		if (g_uiHideHandlerId != 0)
		{
			g_source_remove (g_uiHideHandlerId);
			g_uiHideHandlerId = 0;
		}
		g_iWindowMove = STATIC;
		return FALSE;
	}
}

static gboolean
on_expose (GtkWidget*		pWidget,
	   GdkEventExpose*	pExpose)
{
	gint iWidth;
	gint iHeight;
	cairo_t* pCairoContext = NULL;

	pCairoContext = dock_cairo_create (pWidget->window);
	if (!pCairoContext)
		return FALSE;

	gtk_window_get_size (GTK_WINDOW (pWidget), &iWidth, &iHeight);
	render (pCairoContext, iWidth, iHeight);
	cairo_destroy (pCairoContext);

	return FALSE;
}

static gboolean
on_configure (GtkWidget* pWidget,
	      GdkEventConfigure* pEvent,
	      gpointer userData)
{
	gint iNewWidth = pEvent->width;
	gint iNewHeight = pEvent->height;

	if (iNewWidth != g_iCurrentWidth || iNewHeight != g_iCurrentHeight)
	{
		g_iCurrentWidth = iNewWidth;
		g_iCurrentHeight = iNewHeight;

#ifdef HAVE_GLITZ
		if (g_pGlitzDrawable)
			glitz_drawable_update_size (g_pGlitzDrawable,
						    g_iCurrentWidth,
						    g_iCurrentHeight);
#endif
	}

	return FALSE;
}

static gboolean
on_key_press (GtkWidget*	pWidget,
	      GdkEventKey*	pKey,
	      gpointer		userData)
{
	gint iX = 0;
	gint iY = 0;

	if (pKey->type == GDK_KEY_PRESS)
	{
		switch (pKey->keyval)
		{
			case GDK_Escape :
			case GDK_q :
				gtk_main_quit ();
			break;

			case GDK_Down :
				gtk_window_get_position (GTK_WINDOW (pWidget), &iX, &iY);
				if (iY < g_iScreenHeight - 1)
				{
					iY += 2;
					gtk_window_move (GTK_WINDOW (pWidget), iX, iY);
				}
			break;

			case GDK_Up :
				gtk_window_get_position (GTK_WINDOW (pWidget), &iX, &iY);
				if (iY > g_iScreenHeight - WIN_HEIGHT)
				{
					iY -= 2;
					gtk_window_move (GTK_WINDOW (pWidget), iX, iY);
				}
			break;
		}
	}

	return FALSE;
}

static gboolean
on_button_press (GtkWidget* pWidget,
		 GdkEventButton* pButton,
		 GdkWindowEdge edge)
{
	Icon *pIcon = g_aIcons;

	if (pButton->type == GDK_BUTTON_PRESS)
	{
		switch (pButton->button)
		{
			int icon;

			case 1:
			if (!g_bClicked)
			{
				g_iClickedIcon = -1;
				
				for (icon = 0; icon < SVGS; icon++)
					if (pIcon[icon].fX <= g_fMouseX &&
					    g_fMouseX <= pIcon[icon].fX + pIcon[icon].fWidth * pIcon[icon].fScale)
					{
						g_iClickedIcon = icon;

						g_bClicked = TRUE;
					}

				if (g_bClicked)
				{
					icon = g_iClickedIcon;

					g_fClickedIconX = pIcon[icon].fX
					                + pIcon[icon].fWidth * pIcon[icon].fScale / 2;

					g_uiTimeoutHandlerId = g_timeout_add (30, clicked, (gpointer) pWidget);
					pIcon[icon].bClicked = TRUE;
				}
			}
			break;

			case 2:
			gtk_window_begin_move_drag (GTK_WINDOW (gtk_widget_get_toplevel (pWidget)),
						    pButton->button,
						    pButton->x_root,
						    pButton->y_root,
						    pButton->time);
			break;

			case 3:
			gtk_window_begin_resize_drag (GTK_WINDOW (gtk_widget_get_toplevel (pWidget)),
						      edge,
						      pButton->button,
						      pButton->x_root,
						      pButton->y_root,
						      pButton->time);
			break;
		}
	}

	return FALSE;
}

static gboolean
on_motion_notify (GtkWidget* pWidget,
		  GdkEventMotion* pMotion,
		  gpointer data)
{
	gint iMouseX;
	gint iMouseY;
	gdouble fLastMouseX;

	if (g_bClicked)
		return FALSE;

	gdk_window_get_pointer (pWidget->window, &iMouseX, &iMouseY, NULL);

	fLastMouseX = g_fMouseX;
	g_fMouseX = (gdouble) iMouseX;
	g_fMouseY = (gdouble) iMouseY;

	if (fLastMouseX < g_fMouseX)
		g_fGradientOffsetX = (g_fGradientOffsetX > 0.0f) ?
							 (g_fGradientOffsetX -= 1.0f) :
							 15.0f;
	else if (fLastMouseX > g_fMouseX)
		g_fGradientOffsetX = (g_fGradientOffsetX < 15.0f) ?
							 (g_fGradientOffsetX += 1.0f) :
							 0.0f;

	/* Karl Lattimer - modified
	 * 
	 * There are some hackish bits in here, i'm very tired at present but i think i've cracked it
	 * What I've done here is make it grow when entered and shrink when leaving, this includes
	 * leaving via the sides and entering via the sides.
	 */
	int icon;
	gdouble fStart = iInfluenceGlobal;
	gdouble fEnd = fStart - GAP;
	gdouble fTop = 0;
	gint dir;

	for (icon = 0; icon < SVGS; icon++)
		fEnd += GAP + g_aIcons[icon].fWidth;
	
	if (g_fMagnitude < 0.01)
		g_fMagnitude = 0;
	else if (g_fMagnitude > 1.0)
		g_fMagnitude = 1.0;

	fTop = get_current_dock_offset_y();
	fTop -= fTop * g_fMagnitude;
	fTop -= g_iDockLineWidth;


	// Should get the position of the window and more sensibly detect
	// if dir should be UP when only 3px are showing at the bottom
	// of the screen.
	if (g_fMouseY > (get_current_dock_offset_y() - 9))
		dir = UP;	
	else if (g_fMouseY <= (fTop - 5))
		dir = DOWN;

	if (g_iAtBottom && (g_fMouseY > 1)) {
		dir = UP;
	}
	if ((g_iWindowMove == DOWN) && (dir == UP)) {
		//g_print ("Changing direction DOWN to UP\n");
		if (g_uiHideHandlerId != 0)
		{
			g_source_remove (g_uiHideHandlerId);
			g_uiHideHandlerId = 0;
		}
		
	} else if ((g_iWindowMove == UP) && (dir == DOWN)) {
		//g_print ("Changing direction UP to DOWN\n");
		if (g_uiHideHandlerId != 0)
		{
			g_source_remove (g_uiHideHandlerId);
			g_uiHideHandlerId = 0;
		}
		g_iWindowMove = dir;
	}

	g_iWindowMove = dir;

	if (g_uiHideHandlerId == 0) {
		if (dir == UP) {
			g_uiHideHandlerId = g_timeout_add (15, move_up, (gpointer) pWidget);	
		} else if (dir == DOWN) {
			g_uiHideHandlerId = g_timeout_add (15, move_down, (gpointer) pWidget);
			if (g_uiTimeoutHandlerId != 0)
			{
				g_source_remove (g_uiTimeoutHandlerId);
				g_uiTimeoutHandlerId = 0;
			}
			g_uiTimeoutHandlerId = g_timeout_add (20, shrink_down, (gpointer) pWidget);
		}
	}
	//g_print ("X: %f, Y: %f\n", g_fMouseX, g_fMouseY);

	if ((g_fMouseX > fStart) && (g_fMouseX < fEnd) && (g_fMouseY > fTop)) {
		if ((g_fMagnitude == 0) && (g_uiTimeoutHandlerId == 0)) {
			g_uiTimeoutHandlerId = g_timeout_add (16, grow_up, (gpointer) pWidget);
			if ((g_uiHideHandlerId != 0) && (g_iWindowMove == DOWN)) 
			{
				g_source_remove (g_uiHideHandlerId);
				g_uiHideHandlerId = 0;
			} 
			if (g_uiHideHandlerId == 0) {
				g_uiHideHandlerId = g_timeout_add (15, move_up, (gpointer) pWidget);
			}
		} else if ((g_uiTimeoutHandlerId != 0) && (g_fMagnitude == 1.0)) {
			g_source_remove (g_uiTimeoutHandlerId);
			g_uiTimeoutHandlerId = 0;
			g_fMagnitude = 1.0;
		} else if (g_fMagnitude == 1.0) {
			recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);
			gtk_widget_queue_draw (pWidget);
		}
	} else { 
		if ((g_fMagnitude == 1.0) && (g_uiTimeoutHandlerId == 0)) {
			g_uiTimeoutHandlerId = g_timeout_add (16, shrink_down, (gpointer) pWidget);
		} else if ((g_uiTimeoutHandlerId != 0) && (g_fMagnitude == 0)) {
			g_source_remove (g_uiTimeoutHandlerId);
			g_uiTimeoutHandlerId = 0;
			g_fMagnitude = 0;
		}
	}

	g_uiEvents++;

	return FALSE;
}

static gboolean
on_enter_notify (GtkWidget* pWidget,
		 GdkEventCrossing* pEvent,
		 gpointer data)
{
	if (g_bUseBenchmark)
		g_iAutoMoveDir = 0;

	g_bClicked = FALSE;
	g_fInfluence = INFLUENCE;

	/*if (g_uiHideHandlerId != 0)
	{
		g_source_remove (g_uiHideHandlerId);
		g_uiHideHandlerId = 0;
	}*/

	//g_uiHideHandlerId = g_timeout_add (100, move_up, (gpointer) pWidget);

	return FALSE;
}

static gboolean
benchmark (gpointer data)
{
	g_fMouseX += g_iAutoMoveDir;

	if ((g_iAutoMoveDir > 0 && g_fMouseX >= g_iCurrentWidth) ||
	    (g_iAutoMoveDir < 0 && g_fMouseX <= 0))
		g_iAutoMoveDir = -g_iAutoMoveDir;

	recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);

	gtk_widget_queue_draw ((GtkWidget*) data);

	return TRUE;
}

static gboolean
on_leave_notify (GtkWidget* pWidget,
		 GdkEventCrossing* pEvent,
		 gpointer data)
{
	if (g_uiTimeoutHandlerId != 0)
	{
		g_source_remove (g_uiTimeoutHandlerId);
		g_uiTimeoutHandlerId = 0;
	}
	if (g_uiTimeoutHandlerId != 0)
		return FALSE;

	if (g_bUseBenchmark)
		g_iAutoMoveDir = 1;
	
	g_uiTimeoutHandlerId = g_timeout_add (17, shrink_down, (gpointer) pWidget);

	
	if (g_uiHideHandlerId != 0)
	{
		g_source_remove (g_uiHideHandlerId);
		g_uiHideHandlerId = 0;			
	} 

	g_uiHideHandlerId = g_timeout_add (15, move_down, (gpointer) pWidget);
	return FALSE;
}

static gboolean
none_clicked (void)
{
	int i;

	for (i = 0; i < SVGS; i++)
		if (g_aIcons[i].bClicked)
			return FALSE;

	return TRUE;
}

static gboolean
clicked (gpointer data)
{
	if (g_fInfluence / INFLUENCE > 0.01)
		g_fInfluence *= 0.8;
	else
		g_fInfluence = 0.0;

	g_fMouseX = g_fMouseX * 0.63 + g_fClickedIconX * 0.37;

	recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);

	gtk_widget_queue_draw ((GtkWidget*) data);

	if (g_fInfluence < 0.01 && none_clicked ())
	{
		g_uiTimeoutHandlerId = 0;

		if (g_bUseBenchmark)
		{
			g_iAutoMoveDir = 1;
			g_fInfluence = INFLUENCE;
		}

		return FALSE;
	}
	else
		return TRUE;
}

static gboolean
grow_up (gpointer data)
{
	if (g_fMagnitude < 0.01)
		g_fMagnitude = 0.01;
	
	if (g_fMagnitude < 0.99)
		g_fMagnitude /= 0.5;
	
	if (g_fMagnitude > 1.0)
		g_fMagnitude = 1.0;
				
	recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);
	gtk_widget_queue_draw ((GtkWidget*) data);

	if (g_fMagnitude > 0.99 && none_clicked ())
	{
		g_uiTimeoutHandlerId = 0;
		return FALSE;
	}
	else
		return TRUE;
}

static gboolean
shrink_down (gpointer data)
{
	if (g_fMagnitude > 0.01)
		g_fMagnitude *= 0.618;
	else
		g_fMagnitude = 0.0;
		
	recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);
	gtk_widget_queue_draw ((GtkWidget*) data);

	if (g_fMagnitude < 0.01 && none_clicked ())
	{
		g_fMagnitude = 0;
		g_uiTimeoutHandlerId = 0;
		return FALSE;
	}
	else
		return TRUE;
}

static gdouble
scale (gdouble delta, gdouble fInfluence, gdouble fMagnitude)
{
	gdouble rate = fInfluence / M_PI_2;

	if (delta <= -fInfluence)
		return 1.0;
	if (delta >= +fInfluence)
		return 1.0;

		return 1.0 + fMagnitude * cosf (delta / rate);
}

static gdouble
scaleIntegral (gdouble delta, gdouble fInfluence, gdouble fMagnitude)
{
	gdouble rate = fInfluence / M_PI_2;
	gdouble expansion = fMagnitude * rate;

	if (delta <= -fInfluence)
		return delta - expansion;

	if (delta >= +fInfluence)
		return delta + expansion;

		return delta + expansion * sin (delta / rate);
}

static gdouble
scaledPosition (gdouble fX, gdouble fMouseX, gdouble fInfluence, gdouble fMagnitude)
{
	return fMouseX + scaleIntegral (fX - fMouseX, fInfluence, fMagnitude);
}

static void
recalc_positions (Icon* pIcon, int iIcons,
		  gdouble fMouseX, gdouble fGap, gdouble fInfluence, gdouble fMagnitude)
{
	int icon;
	gdouble fStart = iInfluenceGlobal;
	gdouble fEnd = fStart - fGap;
	gdouble fLastX = fStart;
	gdouble fAdjustment = 0.0;

	for (icon = 0; icon < iIcons; icon++)
		fEnd += fGap + pIcon[icon].fWidth;

	for (icon = 0; icon < iIcons; icon++)
	{
		gdouble fWidth;
		gdouble fHeight;
		gdouble fScale;
		gdouble fStartX;
		gdouble fX;
		gdouble fScaledX;

		fWidth = pIcon[icon].fWidth;
		fHeight = pIcon[icon].fHeight;

		fStartX = fLastX;
		fLastX += fWidth + fGap;

		//fX = MIN (MAX (fStartX, fMouseX), fStartX + fWidth * 0.5);
		fX = fStartX + fWidth * 0.5;
		fScaledX = scaledPosition (fX, fMouseX, fInfluence, fMagnitude);

		fScale = scale (fX - fMouseX, fInfluence, fMagnitude);

		fWidth *= fScale;
		fHeight *= fScale;

		pIcon[icon].fX = fScaledX - (fX - fStartX) * fScale;

		if (pIcon[icon].bClicked)
		{
			pIcon[icon].fY = 50.0 + 2.0 * pIcon[icon].fHeight - pIcon[icon].fScale * pIcon[icon].fHeight - 40.0f * fabs (sin (pIcon[icon].fPeriod));
			pIcon[icon].fPeriod += 3*M_PI/50.0f;

			if (pIcon[icon].fPeriod >= 2 * M_PI)
			{
				GError* pError = NULL;

				pIcon[icon].fPeriod = 0.0f;
				pIcon[icon].bClicked = FALSE;
				g_spawn_command_line_async (pIcon[icon].acCommand, &pError);
			}
		}
		else
			pIcon[icon].fY = 50.0 + 2.0 * pIcon[icon].fHeight - fHeight;

		pIcon[icon].fScale = fScale;
	}

	if (iIcons > 0)
	{
		if (fMouseX < fStart)
			fAdjustment = fStart - pIcon[0].fX;
		else if (fMouseX > fEnd)
			fAdjustment = fEnd - (pIcon[iIcons - 1].fWidth * pIcon[iIcons - 1].fScale) - pIcon[iIcons - 1].fX;
	}

	for (icon = 0; icon < iIcons; icon++)
		pIcon[icon].fX += fAdjustment;
}

static void
fill_icon_buffer (cairo_t* pSourceContext,
		  Icon* pIcon,
		  int iIcons,
		  gdouble fMaxScale)
{
	int icon;

	for (icon = 0; icon < iIcons; icon++)
	{
		cairo_surface_t* pNewSurface;
		cairo_t* pCairoContext;

		pNewSurface = cairo_surface_create_similar (cairo_get_target (pSourceContext),
							    CAIRO_CONTENT_COLOR_ALPHA,
							    ceil (pIcon[icon].fWidth * fMaxScale),
							    ceil (pIcon[icon].fHeight * fMaxScale));
		pCairoContext = cairo_create (pNewSurface);

		/* draw stuff */
		cairo_scale (pCairoContext, fMaxScale, fMaxScale);
		rsvg_handle_render_cairo (pIcon[icon].pHandle, pCairoContext);

		cairo_destroy (pCairoContext);
		pIcon[icon].pIconBuffer = pNewSurface;
	}
}

static void
fill_text_buffer (cairo_t* pSourceContext,
		  Icon* pIcon,
		  int iIcons,
		  gdouble fMaxScale)
{
	// KL fixed the text, made it look decent
	int icon;
	PangoFontDescription *pDesc;
	PangoLayout *pLayout;

	pLayout = pango_cairo_create_layout (pSourceContext);

	pDesc = pango_font_description_new ();
	pango_font_description_set_absolute_size (pDesc, 15 * PANGO_SCALE);
	pango_font_description_set_family_static (pDesc, "sans");
	pango_font_description_set_weight(pDesc, PANGO_WEIGHT_SEMIBOLD);
	pango_layout_set_font_description (pLayout, pDesc);
	pango_font_description_free (pDesc);

	for (icon = 0; icon < iIcons; icon++)
	{
		cairo_surface_t* pNewSurface;
		cairo_t* pCairoContext;
		gint i;
		PangoRectangle ink, log;

		pango_layout_set_text (pLayout, pIcon[icon].acName, -1);
		pango_layout_get_pixel_extents (pLayout, &ink, &log);

		pNewSurface = cairo_surface_create_similar (cairo_get_target (pSourceContext),
							    CAIRO_CONTENT_COLOR_ALPHA,
							    ink.width + 2, ink.height + 2);
		pCairoContext = cairo_create (pNewSurface);
		cairo_translate (pCairoContext, -ink.x, -ink.y);

		cairo_push_group (pCairoContext);
		cairo_set_source_rgb (pCairoContext, 0.2, 0.2, 0.2);
		for (i = 0; i < 4; i++)
		{
			cairo_move_to (pCairoContext, i&2, 2*(i&1));
			pango_cairo_show_layout (pCairoContext, pLayout);
		}
		cairo_pop_group_to_source (pCairoContext);
		cairo_paint_with_alpha (pCairoContext, .7);

		cairo_set_source_rgb (pCairoContext, 1., 1., 1.);
		cairo_move_to (pCairoContext, 1., 1.);
		pango_cairo_show_layout (pCairoContext, pLayout);

		cairo_destroy (pCairoContext);

		/* set_device_offset is buggy, doesn't work for positive
		 * offsets.  so we use explicit offsets... so unfortunate.
		cairo_surface_set_device_offset (pNewSurface, 
						 log.width / 2. - ink.x,
						 log.height     - ink.y);
		 */
		pIcon[icon].fTextXOffset = log.width / 2. - ink.x;
		pIcon[icon].fTextYOffset = log.height     - ink.y;
		pIcon[icon].pTextBuffer = pNewSurface;
	}

	g_object_unref (pLayout);
}

static gboolean
spit_out_redraws (gpointer data)
{
	g_print ("%d fps, %d events\n", g_uiRedraws, g_uiEvents);
	g_uiRedraws = 0;
	g_uiEvents = 0;

	return TRUE;
}

int
main (int argc, char** argv)
{
	GtkWidget* pWindow = NULL;
	GdkScreen *gdkscreen = NULL;
	int icon = 0;
	GError* pError = NULL;
	guint uiHandlerId = 0;
	guint uiBenchmarkId = 0;
	cairo_t* pCairoContext = NULL;
	gint i;
 
 	gtk_init (&argc, &argv);
 
	for (i = 0; i < argc; i++)
	{
		if (strcmp (argv[i], "--benchmark") == 0)
			g_bUseBenchmark = TRUE;
		else if (strcmp (argv[i], "--glitz") == 0)
		{
#ifdef HAVE_GLITZ
			g_bUseGlitz = TRUE;
#else
			fprintf (stderr, "Not compiled with glitz\n");
			return 1;
#endif
		}
		else if (strcmp (argv[i], "--no-glitz") == 0)
		{
#ifdef HAVE_GLITZ
			g_bUseGlitz = FALSE;
#endif
		}
		else if (strcmp (argv[i], "--no-text") == 0)
			g_bUseText = FALSE;
		else if (strcmp (argv[i], "--no-buffering") == 0)
			g_bUseIconBuffering = FALSE;
		else if (strcmp (argv[i], "--no-keep-above") == 0)
			g_bKeepAbove = FALSE;
		else if (strcmp (argv[i], "--no-skip-pager") == 0)
			g_bSkipPager = FALSE;
		else if (strcmp (argv[i], "--no-skip-taskbar") == 0)
			g_bSkipTaskbar = FALSE;
		else if (strcmp (argv[i], "--no-sticky") == 0)
			g_bSticky = FALSE;
		else if (strcmp (argv[i], "--no-background") == 0)
			g_bDrawBackground = FALSE;
		else if (argv[i][0] == '-')
		{
			gboolean help = (strcmp (argv[i], "--help") == 0);
			fprintf (help ? stdout : stderr,
				 "Usage: %s [--benchmark] [--no-glitz] [--no-text] [--no-buffering] [--no-keep-above] [--no-skip-pager] [--no-skip-taskbar] [--no-sticky] [--no-background] [--help]\n",
				 argv[0]);
			return help ? 0 : 1;
		}
	}


	rsvg_init ();

	for (icon = 0; icon < SVGS; icon++)
	{
		g_aIcons[icon].pHandle = rsvg_handle_new_from_file (g_aIcons[icon].acFileName, &pError);
		rsvg_handle_get_dimensions (g_aIcons[icon].pHandle, &g_DimensionData);
		g_aIcons[icon].fWidth = (gdouble) g_DimensionData.width;
		g_aIcons[icon].fHeight = (gdouble) g_DimensionData.height;
		g_aIcons[icon].bClicked = FALSE;
		g_aIcons[icon].fPeriod = 0.0f;
	}

	pWindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gdkscreen = gtk_window_get_screen (GTK_WINDOW (pWindow));
	g_iCurrentWidth = gdk_screen_get_width (gdkscreen);

	g_iWindowWidth = gdk_screen_get_width (gdkscreen);
	g_iWindowHeight = 3 * g_aIcons[0].fHeight;

#ifdef HAVE_GLITZ
	if (g_bUseGlitz)
	{
	    glitz_drawable_format_t templ, *format;
	    unsigned long	    mask = 0;
	    XVisualInfo		    *vinfo = NULL;
	    int			    screen = 0;
	    GdkVisual		    *visual;
	    GdkColormap		    *colormap;
	    GdkDisplay		    *gdkdisplay;
	    Display		    *xdisplay;

	    templ.doublebuffer = 1;
	    mask |= GLITZ_FORMAT_DOUBLEBUFFER_MASK;

	    gdkdisplay = gtk_widget_get_display (pWindow);
	    xdisplay   = gdk_x11_display_get_xdisplay (gdkdisplay);

	    i = 0;
	    do {
		format = glitz_glx_find_window_format (xdisplay, screen,
						       mask, &templ, i++);
		if (format)
		{
		    vinfo = glitz_glx_get_visual_info_from_format (xdisplay,
								   screen,
								   format);
		    if (vinfo->depth == 32)
		    {
			gDrawFormat = format;
			break;
		    }
		    else
		    {
			if (!gDrawFormat)
			    gDrawFormat = format;
		    }
		}
	    } while (format);

		if (!gDrawFormat)
		{
			fprintf (stderr, "no double buffered GLX visual\n");
			return 1;
		}

		vinfo = glitz_glx_get_visual_info_from_format (xdisplay,
							       screen,
							       gDrawFormat);

		visual = gdkx_visual_get (vinfo->visualid);
		colormap = gdk_colormap_new (visual, TRUE);

		gtk_widget_set_colormap (pWindow, colormap);
		gtk_widget_set_double_buffered (pWindow, FALSE);
	}
	else
#endif
	on_alpha_screen_changed (pWindow, NULL, NULL);

	gtk_widget_set_app_paintable (pWindow, TRUE);
	gtk_window_set_decorated (GTK_WINDOW (pWindow), FALSE);
	gtk_window_set_resizable (GTK_WINDOW (pWindow), FALSE);
	gtk_window_set_title (GTK_WINDOW (pWindow), "cairo-dock");
	gtk_widget_set_size_request (pWindow, g_iCurrentWidth, g_iCurrentHeight);
	g_iScreenHeight = gdk_screen_get_height (gdkscreen);
	g_iWindowPositionX = 0;
	g_iWindowPositionY = g_iScreenHeight - WIN_HEIGHT;
	gtk_window_move (GTK_WINDOW (pWindow), g_iWindowPositionX, g_iWindowPositionY);
	gtk_widget_add_events (pWindow,
			       GDK_BUTTON_PRESS_MASK |
			       GDK_POINTER_MOTION_MASK |
			       GDK_POINTER_MOTION_HINT_MASK);

	gtk_widget_show (pWindow);
	if (g_bSticky)
		gtk_window_stick (GTK_WINDOW (pWindow));
	gtk_window_set_keep_above (GTK_WINDOW (pWindow), g_bKeepAbove);
	gtk_window_set_skip_pager_hint (GTK_WINDOW (pWindow), g_bSkipPager);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW (pWindow), g_bSkipTaskbar);

	g_signal_connect (G_OBJECT (pWindow),
			  "destroy",
			  G_CALLBACK (gtk_main_quit),
			  NULL);
	g_signal_connect (G_OBJECT (pWindow),
			  "expose-event",
			  G_CALLBACK (on_expose),
			  NULL);
	g_signal_connect (G_OBJECT (pWindow),
			  "configure-event",
			  G_CALLBACK (on_configure),
			  NULL);
	g_signal_connect (G_OBJECT (pWindow),
			  "key-press-event",
			  G_CALLBACK (on_key_press),
			  NULL);
	g_signal_connect (G_OBJECT (pWindow),
			  "button-press-event",
			  G_CALLBACK (on_button_press),
			  NULL);
	g_signal_connect (G_OBJECT (pWindow),
			  "motion-notify-event",
			  G_CALLBACK (on_motion_notify),
			  NULL);
	g_signal_connect (G_OBJECT (pWindow),
			  "enter-notify-event",
			  G_CALLBACK (on_enter_notify),
			  NULL);
	g_signal_connect (G_OBJECT (pWindow),
			  "leave-notify-event",
			  G_CALLBACK (on_leave_notify),
			  NULL);

	pCairoContext = dock_cairo_create (pWindow->window);
	if (g_bUseText)
		fill_text_buffer (pCairoContext, g_aIcons, SVGS, 2.0);
	if (g_bUseIconBuffering)
		fill_icon_buffer (pCairoContext, g_aIcons, SVGS, 2.0);
	cairo_destroy (pCairoContext);

#ifdef HAVE_GLITZ
	if (g_bUseGlitz)
	{
		glitz_format_t templ;
		GdkDisplay	   *gdkdisplay;
		Display	   *xdisplay;
		Window	   xid;

		gdkdisplay = gdk_display_get_default ();
		xdisplay   = gdk_x11_display_get_xdisplay (gdkdisplay);

		xid = gdk_x11_drawable_get_xid (GDK_DRAWABLE (pWindow->window));

		g_pGlitzDrawable = glitz_glx_create_drawable_for_window (xdisplay,
									 0,
									 gDrawFormat,
									 xid,
									 g_iCurrentWidth,
									 g_iCurrentHeight);
		if (!g_pGlitzDrawable)
		{
			fprintf (stderr, "failed to create drawable\n");
			return 1;
		}

		templ.color        = gDrawFormat->color;
		templ.color.fourcc = GLITZ_FOURCC_RGB;

		g_pGlitzFormat = glitz_find_format (g_pGlitzDrawable,
						    GLITZ_FORMAT_RED_SIZE_MASK   |
						    GLITZ_FORMAT_GREEN_SIZE_MASK |
						    GLITZ_FORMAT_BLUE_SIZE_MASK  |
						    GLITZ_FORMAT_ALPHA_SIZE_MASK |
						    GLITZ_FORMAT_FOURCC_MASK,
						    &templ,
						    0);
		if (!g_pGlitzFormat)
		{
			fprintf (stderr, "couldn't find surface format\n");
			return 1;
		}
	}
#endif

	g_fInfluence = INFLUENCE;
	g_fMagnitude = 0.0;
	
	gdouble fEnd;

	for (icon = 0; icon < SVGS; icon++)
		fEnd += GAP + g_aIcons[icon].fWidth;
	
	iInfluenceGlobal = ( g_iWindowWidth - (int)fEnd)/2;

	recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);

	if (g_bUseBenchmark)
	{
		g_fMagnitude = 1.0;
		recalc_positions (g_aIcons, SVGS, g_fMouseX, GAP, g_fInfluence, g_fMagnitude);

 		uiHandlerId = g_timeout_add (1000, spit_out_redraws, NULL);
		uiBenchmarkId = g_idle_add (benchmark, (gpointer) pWindow);
	}


	gtk_main ();

	if (uiHandlerId)
		g_source_remove (uiHandlerId);
	if (uiBenchmarkId)
		g_source_remove (uiBenchmarkId);

	for (icon = 0; icon < SVGS; icon++)
	{
		rsvg_handle_free (g_aIcons[icon].pHandle);
		cairo_surface_destroy (g_aIcons[icon].pIconBuffer);
	}

	rsvg_term ();

	return 0;
}
